# REST parameters

restServicePort = 7378
restBeaconPort = 4242
restStatePort = 4243
restNotifyPort = 4244
ipv4MulticastAddr = "224.0.0.1"
ipv6MulticastAddr = "ff02::1"
multicastAddr = ipv4MulticastAddr
restBeaconInterval = 10
restHeartbeatInterval = 30
restTimeout = 60
beaconTimeout = 60
restRetryInterval = 10
